
/*
 *	Access control list for htppu
 *	
 *	by Tomi Manninen, OH2BNS
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include "access.h"
#include "log.h"
#include "hmalloc.h"
#include "config.h"

typedef struct access_s {
	unsigned int	network;
	unsigned int	netmask;
	int		mode;
	struct access_s	*next;
} acclist_t;

static struct bitlist_t modelist[] = {
	{ "BAN",	ACC_BAN   },
	{ "USER",	ACC_USER  },
	{ "HOST",	ACC_HOST  },
	{ "RESTRICTED",	ACC_RESTR },
	{ "OBSERVER",	ACC_OBSRV },
	{ NULL,		-1        }
};

static acclist_t *accesslist = NULL;

/*
 *	Convert bits to bitmask
 */

static unsigned long bits_to_mask(int bits)
{
	/* GCC seems to have problems shifting with 32 */
	if (bits == 0)
		return 0L;

        return htonl(0xffffffffL << (32 - bits));
}

/*
 *	Configuration file entry "Access"
 */

int do_access(void *list, int argc, char **argv)
{
	acclist_t *ap;
	struct in_addr inaddr;
	struct hostent *hp;
	char *cp, buf[256];
	
	if (argc < 3)
		return -1;
	
	if ((ap = hcalloc(1, sizeof(acclist_t))) == NULL) {
		/* no memory */
		return -1;
	}
	
	ap->netmask = 0xffffffffL;
	
	if ((cp = strchr(argv[1], '/')) != NULL) {
		*cp = 0;
		ap->netmask = bits_to_mask(atoi(++cp));
	}
	
	if ((hp = gethostbyname(argv[1])) != NULL)
		ap->network = ((struct in_addr *)(hp->h_addr))->s_addr;
	else {
		/* unknown host/network */
		hfree(ap);
		return -1;
	}
	
#ifdef BITCHING_MODE
	if (ap->network & ~(ap->netmask)) {
		/* network address and netmask don't match */
		hfree(ap);
		return -1;
	}
#else
	ap->network &= ap->netmask;
#endif
	
	if ((ap->mode = bitcfg(modelist, argc - 2, &argv[2])) == -1)
		return -1;
		
	ap->next = accesslist;
	accesslist = ap;
	
	inaddr.s_addr = ap->network;
	sprintf(buf, "network: %-15s ", inet_ntoa(inaddr));
	inaddr.s_addr = ap->netmask;
	sprintf(buf + 25, "netmask: %-15s", inet_ntoa(inaddr));
	
	if (ap->mode)
		printbits(ap->mode, modelist, buf, 256);
	else
		strcat(buf, " BAN");
	
	log(L_CONFIG, "Access: %s", buf);
	
	return 0;
}

/*
 *	Check address against ACL
 */

int chk_access(unsigned int addr)
{
	acclist_t *ap = accesslist;
	int ret = ACC_BAN;
	
	if (ap == NULL)
		return ACC_USER | ACC_HOST;
	
	/* Walk the list. Last match counts (actually first in the file) */
	while (ap != NULL) {
		if ((addr & ap->netmask) == ap->network)
			ret = ap->mode;
		ap = ap->next;
	}
	
	return ret;
}
